;;;; tdtd-font.el --- Tony's DTD font lock keywords
;; $Id: tdtd-font.el,v 0.4 1997/12/14 15:41:28 tkg Exp $

;; Copyright (C) 1997 Tony Graham

;; Author: Tony Graham <tgraham@mulberrytech.com>

;;; This file is not part of GNU Emacs.

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License
;; as published by the Free Software Foundation; either version 2
;; of the License, or (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


;;;; Commentary:

;; Font lock mode keywords for use when editing DTDs

;; Send bugs to tgraham@mulberrytech.com


;;;; Code:
(defvar tdtd-font-lock-keywords
  (list
   ;;
   ;; Put comment patterns first so they mask any declarations
   ;; that might be inside the comment
   ;;
   '("<!--[^-]*\\(-[^-]+\\)*-->" . font-lock-comment-face)
   '("--[^-]*\\(-[^-]+\\)*--" . font-lock-comment-face)
   ;;
   ;; XML processing instruction
   ;;
   '("\\(<\\?XML\\)[ \t]+[^?]*\\(\\?[^>]+\\)*\\(\\?>\\)"
     (1 font-lock-keyword-face)
     (3 font-lock-keyword-face))
   ;;
   ;; Lump together attribute, entity, notation, shortref and usemap
   ;; declarations
   ;;
   '("\\(<!\\(ATTLIST\\|ENTITY\\([ \t]+%\\|\\)\\|NOTATION\\|SHORTREF\\|USEMAP\\)\\)[ \t]+\\([^ \t]+\\)[^>]+\\(>\\)"
     (1 font-lock-keyword-face)
     (4 font-lock-variable-name-face)
     (5 font-lock-keyword-face))
   ;;
   ;; Doctype declaration
   ;;
   '("\\(<!DOCTYPE\\)[ \t]+\\([^[]+\\)[ \t]+\\(\\[\\)"
     (1 font-lock-keyword-face)
     (2 font-lock-variable-name-face)
     (3 font-lock-keyword-face))
   ;;
   ;; Implicit Link Specification
   ;;
   '("\\(<!LINKTYPE\\)[ \t]+\\([^ \t]+\\)[ \t]+\\([^ \t]+\\)[ \t]+\\(\\#IMPLIED\\)[ \t\n]+*\\(\\[\\)"
     (1 font-lock-keyword-face)
     (2 font-lock-variable-name-face)
     (3 font-lock-variable-name-face)
     (4 font-lock-variable-name-face)
     (5 font-lock-keyword-face))
   ;;
   ;; Explicit Link Specification
   ;;
   '("\\(<!LINKTYPE\\)[ \t]+\\([^ \t]+\\)[ \t]+\\([^ \t]+\\)[ \t]+\\([^ \t]+\\)[ \t]+*\\(\\[\\)"
     (1 font-lock-keyword-face)
     (2 font-lock-variable-name-face)
     (3 font-lock-variable-name-face)
     (4 font-lock-type-face)
     (5 font-lock-keyword-face))
   ;;
   ;; Declaration subset close and markup delimiter close
   ;;
   '("\\(\\]?\\]\\)[ \t]*\\(>\\)"
     (1 font-lock-keyword-face)
     (2 font-lock-keyword-face))
   ;;
   ;; Marked section start
   ;;
   '("\\(<!\\[\\)[^[]*\\(\\[\\)"
     (1 font-lock-keyword-face)
     (2 font-lock-keyword-face))
   ;;
   ;; Multiple element names in OR group
   ;; Do everything except the element names first
   ;;
   '("\\(<!ELEMENT\\)[ \t]+([^)]+)[ \t]+\\(o[ \t]+o\\|o[ \t]+-\\|-[ \t]+o\\|-[ \t]+-\\|O[ \t]+O\\|O[ \t]+-\\|-[ \t]+O\\)[ \t]+[^>]+\\(>\\)"
     (1 font-lock-keyword-face)
     (2 font-lock-keyword-face)
     (3 font-lock-keyword-face))
   ;;
   ;; Multiple element names in OR group
   ;; Do just the element names second
   ;;
;;   '("<!ELEMENT[ \t]+([ \t]*\\([^ \t|)]+\\)\\([ \t|]+\\([^ \t|)]+\\)+\\)"
;;     (1 font-lock-variable-name-face)
;;     (3 font-lock-variable-name-face))
   ;;
   ;; Multiple element names in OR group
   ;; This catches the last element in an OR group
   ;;
;;   '("<!ELEMENT[ \t]+([ \t]*\\([^ \t|)]+[ \t|]+\\)+\\([^ \t|)]+\\)"
;;     (2 font-lock-variable-name-face))
   ;;
   ;; SGML element name
   ;;
   '("\\(<!ELEMENT\\)[ \t]+\\([^ \t()|]+\\)[ \t]+\\(o[ \t]+o\\|o[ \t]+-\\|-[ \t]+o\\|-[ \t]+-\\|O[ \t]+O\\|O[ \t]+-\\|-[ \t]+O\\)[ \t]+[^>]+\\(>\\)"
     (1 font-lock-keyword-face)
     (2 font-lock-variable-name-face)
     (3 font-lock-keyword-face)
     (4 font-lock-keyword-face))
   ;;
   ;; XML element name
   ;;
   '("\\(<!ELEMENT\\)[ \t]+\\([^ \t()|]+\\)[ \t]+[^>]+\\(>\\)"
     (1 font-lock-keyword-face)
     (2 font-lock-variable-name-face)
     (3 font-lock-keyword-face))
   ;;
   ;; Entity references
   ;;
   '("[%&][^; \t]+;" . font-lock-reference-face)
   ;;
   ;; Put exclusions in a face that will stand out
   ;;
   '("\\(\\(-\\|+\\)(\\)[^)]*\\()\\)"
     (1 font-lock-variable-name-face)
     (3 font-lock-variable-name-face))
   ;;
   ;; Mark the start and end of literals, but don't do anything to their
   ;; contents
   ;;
   '("\\('\\)[^']*\\('\\)"
     (1 font-lock-string-face)
     (2 font-lock-string-face))
   '("\\(\"\\)[^\"]*\\(\"\\)"
     (1 font-lock-string-face)
     (2 font-lock-string-face))
   ;;
   ;; Keywords
   ;; This was created using the make-regexp package then modified by
   ;; hand because some longer strings weren't being matched.
   ;; 
   ;; The SGML keywords we are matching:

   ;; "#" "-//" "+//" "//" "ANY" "APPINFO" "ATTCAP" "ATTCHCAP"
   ;; "ATTCNT" "ATTLIST" "ATTSPLEN" "AVGRPCAP" "BASESET" "BSEQLEN"
   ;; "CAPACITY" "CDATA" "CHANGES" "CHARSET" "CONCUR" "CONREF"
   ;; "CONTROLS" "CURRENT" "DATATAG" "DEFAULT" "DELIMLEN" "DELIM"
   ;; "DESCSET" "DOCTYPE" "DOCUMENT" "DTAGLEN" "DTEMPLEN" "ELEMCAP"
   ;; "ELEMENT" "EMPTY" "ENDTAG" "ENTITIES" "ENTITY" "ENTLVL" "ENTCAP"
   ;; "ENTCHCAP" "EXCLUDE" "EXGRPCAP" "EXNMCAP" "EXPLICIT" "FEATURES"
   ;; "FIXED" "FORMAL" "FUNCHAR" "FUNCTION" "GENERAL" "GRPCNT"
   ;; "GRPGTCNT" "GRPLVL" "ID" "IDCAP" "IDLINK" "IDREF" "IDREFCAP"
   ;; "IDREFS" "IGNORE" "IMPLIED" "IMPLICIT" "INCLUDE" "INITIAL"
   ;; "INSTANCE" "ISO 8879-1986" "ISO 8879:1986" "ISO 8879:1986 (ENR)"
   ;; "LCNMCHAR" "LCNMSTRT" "LINKTYPE" "LINK" "LITLEN" "LKNMCAP"
   ;; "LKSETCAP" "MAPCAP" "MD" "MINIMIZE" "MODEL" "MSICHAR" "MSOCHAR"
   ;; "MSSCHAR" "MS" "NAME" "NAMECASE" "NAMECHAR" "NAMELEN" "NAMES"
   ;; "NAMESTRT" "NAMING" "NDATA" "NET" "NMTOKEN" "NMTOKENS" "NONE"
   ;; "NONSGML" "NOTATION" "NO" "NOTCAP" "NOTCHCAP" "NUMBER" "NUMBERS"
   ;; "NUTOKEN" "NUTOKENS" "OMITTAG" "OTHER" "PCDATA" "PI" "PIC"
   ;; "PILEN" "POSTLINK" "PUBLIC" "PUBLIC" "QUANTITY" "RANK" "RCDATA"
   ;; "RE" "REQUIRED" "RESTORE" "RS" "SCOPE" "SDATA" "SDIF" "SEPCHAR"
   ;; "SEQUENCE" "SGML" "SGMLREF" "SHORTREF" "SHORTTAG" "SHUNCHAR"
   ;; "SIMPLE" "SPACE" "SRCNT" "SRLEN" "STARTTAG" "SUBDOC" "SWITCHES"
   ;; "SYNTAX" "SYSTEM" "TEMP" "TAGLEN" "TAGLVL" "TAB" "TOTALCAP"
   ;; "UCNMCHAR" "UCNMSTRT" "UNUSED" "USELINK" "USEMAP" "VALIDATE"
   ;; "XML" "YES"

   '("\\((ENR)\\|#\\|\\+//\\|-//\\|//\\|A\\(NY\\|PPINFO\\|TT\\(CAP\\|CHCAP\\|CNT\\|LIST\\|SPLEN\\)\\|VGRPCAP\\)\\|B\\(ASESET\\|SEQLEN\\)\\|C\\(APACITY\\|DATA\\|HA\\(NGES\\|RSET\\)\\|ON\\(CUR\\|REF\\|TROLS\\)\\|URRENT\\)\\|D\\(ATATAG\\|E\\(FAULT\\|LIM\\(LEN\\|\\)\\|SCSET\\)\\|OC\\(TYPE\\|UMENT\\)\\|T\\(AGLEN\\|EMPLEN\\)\\)\\|E\\(LEM\\(CAP\\|ENT\\)\\|MPTY\\|N\\(DTAG\\|T\\(IT\\(IES\\|Y\\)\\|LVL\\|C\\(AP\\|HCAP\\)\\)\\)\\|X\\(CLUDE\\|GRPCAP\\|NMCAP\\|PLICIT\\)\\)\\|F\\(EATURES\\|IXED\\|ORMAL\\|UNC\\(HAR\\|TION\\)\\)\\|G\\(ENERAL\\|RP\\(C\\(AP\\|NT\\)\\|GTCNT\\|LVL\\)\\)\\|I\\(D\\(CAP\\|LINK\\|REF\\(CAP\\|S\\)?\\|\\)\\|GNORE\\|MPLI\\(ED\\|CIT\\)\\|N\\(CLUDE\\|ITIAL\\|STANCE\\)\\|SO 8879[-:]1986\\)\\|L\\(CNM\\(CHAR\\|STRT\\)\\|I\\(TLEN\\|NK\\(TYPE\\|\\)\\)\\|K\\(NMCAP\\|SETCAP\\)\\)\\|M\\(APCAP\\|D\\|INIMIZE\\|ODEL\\|S\\(ICHAR\\|OCHAR\\|SCHAR\\|\\)\\)\\|N\\(AM\\(E\\(CASE\\|CHAR\\|S\\(TRT\\)?\\|LEN\\|\\)\\|ING\\)\\|DATA\\|ET\\|MTOKENS?\\|O\\(N\\(E\\|SGML\\)\\|T\\(\\(CH\\)?CAP\\|ATION\\)\\|\\)\\|U\\(MBERS?\\|TOKENS?\\)\\)\\|O\\(MITTAG\\|THER\\)\\|P\\(CDATA\\|I\\(C\\|LEN\\)?\\|OSTLINK\\|UBLIC\\)\\|QUANTITY\\|R\\(S\\|ANK\\|CDATA\\|E\\(QUIRED\\|STORE\\|\\)\\)\\|S\\(COPE\\|D\\(ATA\\|IF\\)\\|E\\(PCHAR\\|QUENCE\\)\\|GML\\(REF\\|\\)\\|H\\(ORT\\(REF\\|TAG\\)\\|UNCHAR\\)\\|IMPLE\\|PACE\\|R\\(CNT\\|LEN\\)\\|TARTTAG\\|UBDOC\\|WITCHES\\|Y\\(NTAX\\|STEM\\)\\)\\|T\\(A\\(B\\|GL\\(EN\\|VL\\)\\)\\|EMP\\|OTALCAP\\)\\|U\\(CNM\\(CHAR\\|STRT\\)\\|NUSED\\|SE\\(LINK\\|MAP\\)\\)\\|VALIDATE\\|XML\\|YES\\)" . font-lock-type-face)
   ;;
   ;;
   ;; Connectors
   ;;
   '("[,()|&]" . font-lock-function-name-face)
   ;;
   ;; Occurence indicators
   ;;
   '("[+*?]" . font-lock-string-face)
   ;;
   )
  "Additional expressions to highlight in SGML mode.")

(defun tdtd-font-lock-mark-block-function ()
  "Function to mark the area of text we want to fontify with
font-lock-fontify-block.  Set font-lock-mark-block-function to
this function for this function to take effect.

This function marks the area beginning five \"<!\" before point and five
\">\" at ends of lines after point.  The default without a function like
this is to fontify 16 lines before and after point, but then the region
often starts or ends partway through a comment or declaration, turning
that half white because the keywords didn't match, and it just looks so
ugly."
  (let ((current-point (point)))
    (re-search-forward ">$" (point-max) 'limit 5)
    (set-mark (point))
    (goto-char current-point)
    (re-search-backward "^<!" (point-min) 'limit 5)))

(provide 'tdtd-font)
