;;;; tdtd.el --- Tony's DTD formatting macros
;; $Id: tdtd.el,v 1.9 1997/12/14 15:44:12 tkg Exp $

;; Copyright (C) 1996, 1997 Tony Graham

;; Author: Tony Graham <tgraham@mulberrytech.com>

;;; This file is not part of GNU Emacs.

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License
;; as published by the Free Software Foundation; either version 2
;; of the License, or (at your option) any later version.
;; 
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;; 
;; You should have received a copy of the GNU General Public License
;; along with this program; if not, write to the Free Software
;; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


;;;; Commentary:

;; Macros for editing DTDs

;; Send bugs to tgraham@mulberrytech.com


;;;; Code:
(provide 'tdtd)

(setq tdtd-dtd-max-column 70
      tdtd-mdc-max-column (1- tdtd-dtd-max-column)
      tdtd-comment-start-column 24
      tdtd-comment-max-column (- tdtd-dtd-max-column 3)
      tdtd-sgml-comment-max-column (- tdtd-dtd-max-column 2)
      tdtd-attribute-name-column 16
      tdtd-attribute-comment-open-column 12
      tdtd-attribute-comment-start-column 32
      tdtd-declared-value-column 24
      tdtd-default-value-column 48
      tdtd-element-name-column 11
      tdtd-element-tag-omission-column 24
      tdtd-element-content-model-start-column 28
      tdtd-element-content-model-continuation-column 29)

(setq tdtd-line-comment
      "<!-- ============================================================= -->\n")

(setq tdtd-header-comment-fill-prefix "<!--             ")

(setq tdtd-sgml-mdo "<!")
(setq tdtd-sgml-mdc ">")

(setq tdtd-sgml-comment-start "<!--")
(setq tdtd-sgml-comment-end "-->")

(setq tdtd-xml-comment-start "<!--")
(setq tdtd-xml-comment-start-regexp "<!--")
(setq tdtd-xml-comment-end "-->")

;; SGML Syntactic Literals
(setq tdtd-empty-literal "EMPTY")

;; If you want to untabify everything every time, add this to functions
;;  (untabify (point-min) (point-max))
;; Its easier to just not insert tabs using:
;;  (setq indent-tabs-mode nil)
(defun tdtd-untabify-all ()
  "Untabify the entire file"
  (interactive)
  (untabify (point-min) (point-max)))

(defun tdtd-fix-entities ()
  "Quick and dirty addition of \";\" to entity references lacking it"
  (interactive)
  (save-excursion
    (goto-char (point-min))
    (while
	(re-search-forward "\\([%&][A-Za-z0-9.]+\\)\\([^A-Za-z0-9.;]\\)" nil t)
      (replace-match "\\1;\\2" nil nil))))

(defun tdtd-indent-or-newline-to (target-column)
  "Indent to target column, or if at or past target, insert newline and indent"
  (if (>= (current-column)
	  target-column)
      (insert "\n"))
  (indent-to target-column))

(defun tdtd-center-comment (comment)
  "Center a comment on line"
  (interactive "sComment: ")
  (insert tdtd-sgml-comment-start)
  (indent-to (/ (- tdtd-dtd-max-column (length comment)) 2))
  (insert comment)
  (indent-to tdtd-comment-max-column)
  (insert tdtd-sgml-comment-start)
  (insert "\n"))

(defun tdtd-recenter-comment ()
  "Recenter text in a presumably modified comment line"
  (interactive)
  (save-excursion
    (save-match-data
      (beginning-of-line)
      (if (looking-at "^<!--\\s-*\\(\\w+\\(\\W+\\w+\\)*\\)\\s-*-->$")
	  (let ((contents (match-string 1)))
	    (delete-region (match-beginning 0) (match-end 0))
	    (tdtd-center-comment contents))))))

;; TDTD house style puts all comments starting on a favourite column
(defun tdtd-comment (comment)
  "Insert a comment starting at the usual column"
  (interactive "sComment: ")
  (insert "\n")
  (backward-char)
  (let ((fill-column (- tdtd-comment-max-column 1))
	(fill-prefix "                        ")
	(comment-start tdtd-header-comment-fill-prefix)
	(saved-auto-fill-function auto-fill-function))
    (auto-fill-mode 1)
    (insert tdtd-sgml-comment-start)
    (indent-to tdtd-comment-start-column)
    (fill-region (point) (save-excursion
			   (insert comment)
			   (point))
		 nil
		 1
		 1)
    ;; The fill does the right thing, but it always ends with
    ;; an extra newline, so we delete the newline.
    (delete-backward-char 1)
    (if (not saved-auto-fill-function)
	(auto-fill-mode 0))
    (indent-to tdtd-comment-max-column)
    (insert tdtd-sgml-comment-end)))

(defun tdtd-big-comment (comment)
  "Insert a comment on line, with <!-- === --> comments above and below"
  (interactive "sComment: ")
  (insert tdtd-line-comment)
  (tdtd-comment comment)
  (insert "\n")
  (insert tdtd-line-comment))

(defun tdtd-recomment ()
  "Fix text position in a presumably modified comment line"
  (interactive)
  (save-excursion
    (save-match-data
      (beginning-of-line)
      (if (looking-at "^\\(<!--\\*\\)?\\s-*\\([^ \t\n>]+\\([- \t]+[^- \t\n>]+\\)*\\)\\s-*\\(\\*-->\\)?$")
	  (let ((contents (match-string 2)))
	    (delete-region (match-beginning 0) (match-end 0))
	    (tdtd-comment contents))))))

;; doesn't work yet.
(defun tdtd-join-comments ()
  "Join two comment lines"
  (interactive)
  (save-excursion
    (save-match-data
      (beginning-of-line)
      (if (looking-at "^\\(<!--\\)?\\s-*\\([^ \t\n>]+\\([- \t]+[^- \t\n>]+\\)*\\)\\s-*\\(-->\\)?$")
	  (let ((contents (match-string 2)))
	    (delete-region (match-beginning 0) (match-end 0))
	    (tdtd-comment contents))))))

(defun tdtd-declare-attribute (attribute-name declared-value default-value comment)
  "Insert an attribute declaration at the current point"
  (interactive "sAttribute name: \nsDeclared value: \nsDefault value: \nsComment: ")
  (let ((fill-column (- tdtd-comment-max-column 1))
	(fill-prefix "                                "))
    (auto-fill-mode 'true)
    (indent-to tdtd-attribute-comment-open-column)
    (insert "--")
    (indent-to tdtd-attribute-comment-start-column)
    (fill-region (point) (save-excursion
			   (insert comment)
			   (point))
		 nil
		 'true)
    ;; The fill does the right thing, but it always ends with
    ;; an extra newline, so we delete the newline.
    (delete-backward-char 1)
    (tdtd-indent-or-newline-to tdtd-comment-max-column)
    (insert "--\n")
    (auto-fill-mode nil)
    (indent-to tdtd-attribute-name-column)
    (insert attribute-name)
    (tdtd-indent-or-newline-to tdtd-declared-value-column)
    (insert declared-value)
    (tdtd-indent-or-newline-to tdtd-default-value-column)
    (insert default-value)))

(defun tdtd-declare-element (element-name element-comment element-tag content-model)
  "Insert an element declaration at the current point"
  (interactive "sElement name: \nsElement Comment: \nsElement Tag: \nsContent Model: ")
  (tdtd-comment element-name)
  (insert "\n")
  (if (not
       (string-equal element-comment ""))
      (progn
	(tdtd-comment element-comment)
	(insert "\n")))
  (insert "\n")
  (backward-char)
  (insert "<!ELEMENT  ")
  (insert element-tag)
  (tdtd-indent-or-newline-to tdtd-element-tag-omission-column)
  (if (string-equal content-model tdtd-empty-literal)
      (insert "- o ")
    (insert "- - "))
  (let ((fill-column tdtd-mdc-max-column)
	(fill-prefix "                             ")
	(saved-auto-fill-function auto-fill-function))
    (auto-fill-mode 1)
    (fill-region (point) (save-excursion
			   (insert content-model)
			   (point))
		 nil
		 1
		 1)
    ;; The fill does the right thing, but it always ends with
    ;; an extra newline, so we delete the newline.
    (delete-backward-char 1)
    (if (not saved-auto-fill-function)
	(auto-fill-mode 0))
    (tdtd-indent-or-newline-to tdtd-mdc-max-column)
    (insert ">\n")
    (auto-fill-mode nil)))

(defun tdtd-insert-mdc ()
  "Without moving point, indent to usual column and insert markup
declaration close (>)"
  (interactive)
  (save-excursion
    (let ((current-point (point))
	  (mdc-column 0)
	  (mdc tdtd-sgml-mdc))
      (if (if (looking-at tdtd-sgml-mdo)
	      1
	    (re-search-backward "^<!"))
	  (progn
	    (cond
	     ((looking-at tdtd-xml-comment-start-regexp)
	      (progn
		(setq mdc tdtd-xml-comment-end)
		(setq mdc-column tdtd-comment-max-column)))
	     ((looking-at tdtd-sgml-comment-start)
	      (progn
		(setq mdc tdtd-sgml-comment-end)
		(setq mdc-column tdtd-sgml-comment-max-column)))
	     (t
	      (progn
		(setq mdc tdtd-sgml-mdc)
		(setq mdc-column tdtd-mdc-max-column))))
	    (goto-char current-point)
	    (beginning-of-line)
	    (cond
	     ;; The main "looking-at" regular expression doesn't handle
	     ;; line containing just "-->" or ">" very well, so handle
	     ;; as special case.
	     ((looking-at "[ \t]*\\(--\\)?>$")
	      (progn
	       ;; used to use (kill-line) to kill to end of line, but it
	       ;; gave wrong behaviour when only whitespace to end of line.
	       (kill-region (point) (save-excursion
				      (end-of-line)
				      (point)))
	       (indent-to mdc-column)
	       (insert mdc)))
	     ;; This is the main "looking-at" regular expression that
	     ;; handles most things
	     ((looking-at "^\\(\\(<!\\|[ \t]*\\)?\\(--\\)?[ \t]*\\([^ \t\n>]+\\([- \t]+[^- \t\n>]+\\)*\\)?\\)[ \t]*\\(\\(--\\)?>?\\)?$")
	      (progn
		;;(message ":%s:%s:%s:%s:" mdc mdc-column (match-string 4) (match-string 1))
		(goto-char (match-end 1))
		;; used to use (kill-line) to kill to end of line, but it
		;; gave wrong behaviour when only whitespace to end of line.
		(kill-region (point) (save-excursion
				       (end-of-line)
				       (point)))
		(indent-to mdc-column)
		(insert mdc))))
	    (font-lock-fontify-block))))))

(defun tdtd-insert-end-com ()
  "Indent to usual column and insert comment end delimiter (--)"
  (interactive)
  (or (eolp)
      (kill-line))
  (indent-to tdtd-comment-max-column)
  (insert "--"))

;; A work in progress
(defun tdtd-fill-paragraph (&optional justify)
  "TDTD fill paragraph function"
  (interactive "P")
  ;;(insert "TDTD")
  (save-excursion
    (let ((current-point (point)))
      (if (not (looking-at tdtd-sgml-mdo))
	  (re-search-backward "^<!"))
      (cond
       ((looking-at "\\(<!ELEMENT\\)")
	(goto-char (match-end 1))
	(if (looking-at "[ \t\r\n]+")
	    (kill-region (match-beginning 0) (match-end 0)))
	(tdtd-indent-or-newline-to tdtd-element-name-column)
	(looking-at "")
	t)
       ;; If we're not doing anything, return nil so the built-in function
       ;; will run
       (t nil)))))

(defun tdtd-comment-setup ()
  "Setup the auto-fill variables for comments in the body of the DTD"
  (interactive)
  (setq fill-prefix "                        ")
  (auto-fill-mode 'true)
  (setq fill-column 69))

(defun tdtd-design-comment-setup ()
  "Setup the auto-fill variables for \"Design Considerations\" comments"
  (interactive)
  (setq fill-prefix "      ")
  (auto-fill-mode 'true)
  (setq fill-column 69))

(defun tdtd-declaration-setup ()
  "Setup the auto-fill variables for declarations"
  (interactive)
  (setq fill-prefix "                             ")
  (auto-fill-mode 'true)
  (setq fill-column 69))

(defun tdtd-init-comment-setup ()
  "Setup the auto-fill variables for initial comments"
  (interactive)
  (setq fill-prefix "<!--            ")
  (auto-fill-mode 'true)
  (setq fill-column 66))

;;(setq fill-paragraph-function 'tdtd-fill-paragraph)

(define-key dtd-mode-map "\C-c>"
  'tdtd-insert-mdc)

(define-key dtd-mode-map "\C-c\C-e"
  'tdtd-declare-element)

(define-key dtd-mode-map "\C-cc"
  'tdtd-comment)

(define-key dtd-mode-map "\C-cb"
  'tdtd-big-comment)

(define-key dtd-mode-map "\M-g\M-f"
  'font-lock-fontify-buffer)

(defun tdtd-startup ()
  "Functions to perform when starting with TDTD stuff"
  (interactive)
  (setq auto-fill-function 'nil))

;; Not quite sure what this gets me, but these variables seem to apply
;; in auto-fill mode
(make-local-variable 'comment-start)
(setq comment-start "<!-- ")
(make-local-variable 'comment-end)
(setq comment-end " -->")

;;;; Last provisions
(provide 'tdtd)
;;;(provide 'dtd-mode)

;;; tdtd.el ends here
